/****************************************************************************/
/** G-FIT V1.0                                                             **/
/** Fitting hyper-Erlang distributions to trace data with the EM algorithm **/
/**                                                                        **/
/** Developed by Axel Thuemmler, July/August 2004                          **/
/** Copyright (c) 2004 University of Dortmund. All rights reserved.        **/
/**                                                                        **/
/** The G-FIT software package is distributed in open-source under the     **/
/** terms of the GNU general public license agreement. To document the     **/
/** usage of G-FIT, we ask each user to cite our scientific publications   **/
/** mentioned on our Web site.                                             **/
/****************************************************************************/
//-------------------------------------
// Implementation of class TGerlangFit
//-------------------------------------

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cmath>
#include <cfloat>

#include "gerlangfit.h"


// write a file with convergence informations
#define CONVFILE			"./curves/conv.dat"

// write a file with logging informations
//#define G_LOGFILE			"./gfit.log"


using namespace std;


//-------------
// Constructor
//-------------
TGerlangFit::TGerlangFit(TTrace* _trace, double _eps, TConvCheck _conv_check)
{

#ifdef G_LOGFILE
	glog = new ofstream(G_LOGFILE);
	glog->precision(16);
	*glog << "---------------------------------------------------------- START LOGFILE" << endl;
#endif

#ifdef CONVFILE
	conv = new ofstream(CONVFILE);
#endif

	count = 0;
	eps = _eps;
	conv_check = _conv_check;

	max_diff = 0.0;
	trace = _trace;

	lptr = new double*[2];
	wptr = new double*[2];

	tmp1 = new double[trace->size];
	tmp2 = NULL;
}


//------------
// Destructor
//------------
TGerlangFit::~TGerlangFit()
{

#ifdef G_LOGFILE
	*glog << "------------------------------------------------------------ END LOGFILE" << endl;
	glog->close();
	delete glog;
#endif

#ifdef CONVFILE
	conv->close();
	delete conv;
#endif

	delete [] lptr;
	delete [] wptr;

	delete [] tmp1;
}


double TGerlangFit::get_max_diff()
{
	return max_diff;
}

unsigned long TGerlangFit::get_count()
{
	return count;
}


//-------------------------------------------------------
// Fitting a hyper-Erlang distribution to the trace data
//-------------------------------------------------------
void TGerlangFit::fit(TGerlangDist* gdist)
{

	// allocation of temporary memory for pre-computed value (allows faster iterations)
	tmp2 = new double[trace->size*gdist->m];

	fit_base(gdist);
	
	delete [] tmp2;
}


//-------------------------------------------------------------------
// Fit all possible hyper-Erlang distributions (HErD) with N states
// to the trace and return the best fitted distribution according to
// the log-likelihood value
//   gdist = contains the best HErD found
//   N     = overall number of states
//   m_min = consider all HErD with at least m_min Erlang branches
//   m_max = consider all HErD with at most m_max Erlang branches
//           (1 <= m_min <= m_max <= N)
//-------------------------------------------------------------------
void TGerlangFit::fit_all(TGerlangDist* gdist, unsigned int N, unsigned int m_min, unsigned int m_max)
{

	// allocation of temporary memory for pre-computed value (allows faster iterations)
	tmp2 = new double[trace->size*m_max];

	// queue with capacity one for only the best fitted distribution
	queue = new TQueue(1, m_max);

	fit_all_base(gdist, N, m_min, m_max);

	delete queue;
	delete [] tmp2;
}


//-------------------------------------------------------------------
// Fit all possible hyper-Erlang distributions (HErD) with N states
// to the trace and return the best fitted distribution according to
// the log-likelihood value. Algorithm applies a priority queueing
// approach with 3 rounds.
//   gdist = contains the best HErD found
//   N     = overall number of states
//   m_min = consider all HErD with at least m_min Erlang branches
//   m_max = consider all HErD with at most m_max Erlang branches
//           (1 <= m_min <= m_max <= N)
//-------------------------------------------------------------------
void TGerlangFit::fit_all_steps(TGerlangDist* gdist, unsigned int N, unsigned int m_min, unsigned int m_max)
{

	// allocation of temporary memory for pre-computed value (allows faster iterations)
	tmp2 = new double[trace->size*m_max];

	unsigned int qs;
	double eps_tmp = eps;
	
	unsigned int rounds = 3;
	double _eps[3];

	if (conv_check == PAR_DIFF)
	{
		// epsilon values for convergence according to parameter difference
		_eps[0] = 1e-2;
		_eps[1] = 1e-3;
		_eps[2] = 1e-4;
	}
	else
	{
		// epsilon values for convergence according to log-likelihood difference
		_eps[0] = 1e-3;
		_eps[1] = 1e-4;
		_eps[2] = 1e-6;
	}

	cout << "starting round no. 1" << endl;	

	// queue the best 25% of all settings but at least 5 and at most 50 settings
	qs = get_num_settings(N, m_min, m_max) / 4;
	if (qs < 5) qs = 5;
	if (qs > 50) qs = 50;

	//initialize priority queue
	queue = new TQueue(qs, m_max);

	// eps for first round
	eps = _eps[0];

	// fit all settings in the first round
	fit_all_base(gdist, N, m_min, m_max);

	// get number of elements in queue
	qs = queue->get_size();

	queue->print(qs);


	for (unsigned int i = 1; i < rounds; i++)
	{
		cout << "starting round no. " << i+1 << endl;

		// eps for this round
		eps = _eps[i];

		for (unsigned int j = 0; j < qs; j++)
		{
			// get the j-th HErD out of the queue
			queue->get(j, gdist);

			cout << "continuing gfit for Erlang phases of length ";
			for (unsigned int k = 0; k < gdist->m-1; k++) cout << gdist->r[k] << ", ";
			cout << gdist->r[gdist->m-1] << endl;
			
			// continue fitting the HErD until new eps is reached
			fit_base(gdist);

			// update the j-th HErD with the new values and possibly move it in the queue
			queue->update(j, gdist->compute_loglikelihood(trace), gdist->lambdas, gdist->weights);			
		}

		// consider in the next round only the best 50%
		qs /= 2;
		if (qs == 0) qs = 1;

//		qs = 1;

		queue->print(qs);
	}

	// get the first element from the final queue
	queue->get_bestfit(gdist);

	eps = eps_tmp;

	delete queue;
	delete [] tmp2;
}


//-------------------------------------------------------------------------
// Calculate number of possibilities to chose the number of Erlang
// branches m and the number of phases r_1,...,r_m of the hyper-Erlang
// distribution for a given overall number of states N.
//------------------------------------------------------------------------
unsigned long TGerlangFit::get_num_settings(unsigned int N, unsigned int m_min, unsigned int m_max)
{
	unsigned int num = 0;

	for (unsigned int m = m_min; m <= m_max; m++)
	{
		num += calc_num_settings(m, N, 1);
	}

	return num;
}


//-----------------------------------------------------
// Compute contribution of Erlang branch i to the
// mean of the overall hyper-Erlang distribution
//-----------------------------------------------------
double TGerlangFit::contrib(TGerlangDist* gdist, unsigned int i)
{
	return gdist->weights[i] * gdist->r[i] / gdist->lambdas[i];
}


int compare_uint(const void* a, const void* b)
{
	return (*(unsigned int*)a - *(unsigned int*)b);
}


//------------------------------------------------------------------
// Fit with adaptive adjustment of phase length of the hyper-Erlang
// distributions
//   steps = number of adaptation steps
//------------------------------------------------------------------
void TGerlangFit::fit_adaptive(TGerlangDist* gdist, unsigned int steps)
{
	unsigned int m = gdist->m;

	// allocation of temporary memory for pre-computed value (allows faster iterations)
	tmp2 = new double[trace->size*m];

	unsigned int* r = new unsigned int[m];
	for (unsigned int i = 0; i < m; i++)
	{
		r[i] = gdist->r[i];
	}

	bool finished = false;	
	unsigned int min, max, r_min, r_max, sec_max;
	double mean;
	unsigned int adap = 0;

	while ((adap < steps) && (!finished))
	{

		adap++;

		qsort(r, m, sizeof(unsigned int), compare_uint);

		// initialize r-values (lambdas and weights are initialized with standard values)
		gdist->init(r);

		// fit the distribution
		fit_base(gdist);

		// for all adaptive steps except the last
		if (adap == steps) continue;

		cout << "log-likelihood: " << gdist->compute_loglikelihood(trace) << endl;
		gdist->print_r();

		min = max = r_min = r_max = sec_max = 0;
		mean = 0.0;

		// find Erlang branch with maximal and minimal contribution to the overall distribution
		for (unsigned int i = 0; i < m; i++)
		{
			cout << "Contribution of Erlang branch " << i+1 << ": " << contrib(gdist, i) << endl;
			if (contrib(gdist, i) > contrib(gdist, max)) max = i;
			if (contrib(gdist, i) < contrib(gdist, min)) min = i;
			if (r[i] > r[r_max]) r_max = i;
			if (r[i] < r[r_min]) r_min = i;
			mean += contrib(gdist, i);
		}

		if (max == 0) sec_max = 1;
			
		// find Erlang branch with second maximal contribution
		for (unsigned int i = 0; i < m; i++)
		{
			if ((i != max) && (contrib(gdist, i) > contrib(gdist, sec_max))) sec_max = i;
		}


		if (contrib(gdist, min)/mean > 0.05)
		{
			finished = true;
			continue;
		}

		// adapt r-values appropriately
		if (r[max] == r[r_max])
		{
			// add 25% of maximal r-value
			r[min] = r[r_max] + (unsigned int) r[r_max] / 4;

			// add at least one phase
			if (r[min] == r[r_max]) r[min]++;
		}
		else if (r[max] == r[r_min])
		{
			// subb 25% of minimal r_value
			r[min] = r[r_min] - (unsigned int) r[r_min] / 4;

			// sub at least one phase
			if (r[min] == r[r_min]) r[min]--;

			// r_min should be at least 1
			if (r[min] < 1) r[min] = 1;
		}
		else 
		{
			r[min] = (r[max] + r[sec_max]) / 2;
		}			
	}

	delete [] r;
	delete [] tmp2;
}


//****************** PRIVATE METHODS **********************************


//-----------------------------------------------------------------------
// Basic implementation of fit_all
//   gdist = contains the best HErD found
//   N     = overall number of states
//   m_min = consider all HErD with at least m_min Erlang branches
//   m_max = consider all HErD with at most m_max Erlang alternatrive
//           (1 <= m_min <= m_max <= N)
//-----------------------------------------------------------------------
void TGerlangFit::fit_all_base(TGerlangDist* gdist, unsigned int N, unsigned int m_min, unsigned int m_max)
{

	unsigned int* r = new unsigned int[m_max];

	// do the fitting for m_min to m_max Erlang branches
	for (unsigned int m = m_min; m <= m_max; m++)
	{
		gdist->adjust_m(m);

		cout << "m = " << m << ":" << endl;

		fit_settings(m, N, 1, r, 0, gdist);
	}

	// get the first element from the queue
	queue->get_bestfit(gdist);

#ifdef G_LOGFILE
	*glog << "final r-vector:" << endl;
	gdist->print_r();
	*glog << endl << "final lambda vector:" << endl;
	gdist->print_lambdas();
	*glog << endl << "final starting probabilities:" << endl;
	gdist->print_weights();
	*glog << endl;
#endif

	delete [] r;
}

//-------------------------------------------------------------------------
// Basic HErD fitting algorithm.
// Extension of EM-algorithm proposed by Haverkort et.al. in
// "Fitting world-wide web request traces with the EM-algorithm",
// published in Performance Evaluation 52 (2003), pp. 175-191
//-------------------------------------------------------------------------
void TGerlangFit::fit_base(TGerlangDist* gdist)
{

	// Step (1): Set initial values
	int old_ptr = 0;
	int new_ptr = 1;

	lptr[old_ptr] = gdist->lambdas;
	wptr[old_ptr] = gdist->weights;
	lptr[new_ptr] = new double[gdist->m];
	wptr[new_ptr] = new double[gdist->m];
	
	double llhptr[2];
	llhptr[old_ptr] = 0.0;

#ifdef G_LOGFILE
	*glog << endl << "-------------------------------------------------------" << endl;
	*glog << "Number of erlang distributions: " << gdist->m << endl;
	*glog << "Number of phases of each erlang distribution: ";
	for (unsigned int i = 0; i < gdist->m-1; i++) *glog << gdist->r[i] << ", ";
	*glog << gdist->r[gdist->m-1] << endl;
	*glog << "-------------------------------------------------------" << endl;
#endif

	double sum1, sum2, p;

	cout << "iteration of EM-algorithm is running ..." << endl;

	do
	{
		count++;

#ifdef G_LOGFILE
		*glog << endl << "Iteration No. " << count << endl;
#endif

		// Step (2): Compute new lambda and initial probabilities

		// Pre-computation of p(x_n | alpha) for all n
		llhptr[new_ptr] = 0.0;
		for (unsigned long n = 0; n < trace->size; n++)
		{
			// initialize tmp1 with DBL_MIN to avoid division by zero
			tmp1[n] = DBL_MIN;
			for (unsigned int i = 0; i < gdist->m; i++)
			{
				p = wptr[old_ptr][i] * gdist->erlang_pdf(gdist->r[i], lptr[old_ptr][i], trace->trace[n]);
				tmp1[n] += p;
				tmp2[i*trace->size+n] = p;
			}
			if (conv_check == LLH_DIFF) llhptr[new_ptr] += log(tmp1[n]);
		}


#ifdef G_LOGFILE
		if (conv_check != LLH_DIFF)
		{
			for (unsigned long n = 0; n < trace->size; n++) llhptr[new_ptr] += log(tmp1[n]);
		}
		*glog << "log-likelihood (before iteration): " << llhptr[new_ptr] << endl;
#endif

		// Sub-steps (a) - (c)
		for (unsigned int i = 0; i < gdist->m; i++)
		{
			sum1 = 0.0;
			sum2 = DBL_MIN;
			for (unsigned long n = 0; n < trace->size; n++)
			{
				// Step (a): Computation of p(i | x_n, lambda_i)
				p = tmp2[i*trace->size+n] / tmp1[n];

				// Computation of sums required for steps (b) and (c)
				sum1 += p;
				sum2 += p*trace->trace[n];
			}

			// Step (b)
			wptr[new_ptr][i] = sum1 / (double) trace->size;

			// Step (c)
			lptr[new_ptr][i] = gdist->r[i] * sum1 / sum2;
			
#ifdef G_LOGFILE
			*glog << "lambda_" << i+1 << " = " << lptr[new_ptr][i] << endl;
			*glog << "weight_" << i+1 << " = " << wptr[new_ptr][i] << endl;
//			*glog << "(sum1 = " << sum1 << ", sum2 = " << sum2 << ")" << endl;
#endif
		}

		// Step (3): Exchange old and new values and calculate stopping criterion
		old_ptr = (old_ptr+1) % 2;
		new_ptr = (new_ptr+1) % 2;

		if (conv_check == PAR_DIFF)
		{
			// compute the maximal difference of the lambdas and weights
			max_diff = 0.0;
			for (unsigned int i = 0; i < gdist->m; i++)
			{
				p = fabs(wptr[new_ptr][i] - wptr[old_ptr][i]);
				if (p > max_diff) max_diff = p;
				p = fabs(lptr[new_ptr][i] - lptr[old_ptr][i]);
				if (p > max_diff) max_diff = p;
			}
		}
		else
		{
			// compute the relative difference of the log-likelihood values
			max_diff = fabs((llhptr[old_ptr] - llhptr[new_ptr]) / llhptr[old_ptr]);
		}

		// show iteration progress
		if ((count % 10) == 0) cerr << "#";
//		if ((count % 10) == 0) cerr << max_diff;

#ifdef CONVFILE
		*conv << count << "  " << max_diff << endl;
#endif

	} while (max_diff > eps);

	cout << endl;

	// put resulting lambda-values and weights in gdist
	gdist->lambdas = lptr[old_ptr];
	gdist->weights = wptr[old_ptr];

	delete [] lptr[new_ptr];
	delete [] wptr[new_ptr];

}


//---------------------------------------------------------------------------
// For a given number m of Erlang branches fit hyper-Erlang distribution
// for all settings of k <= r_1 <= r_2 <= ... <= r_m to the trace
//---------------------------------------------------------------------------
void TGerlangFit::fit_settings(unsigned int m, unsigned int N, unsigned int k, unsigned int* r, unsigned int pos, TGerlangDist* gdist)
{
	if (m == 1)
	{
		r[pos] = N;
		cout << "starting gfit for Erlang phases of length ";

		// print the setting of r_i's
		for (unsigned int i = 0; i < pos; i++) cout << r[i] << ", ";
		cout << N << endl;

		// initialize r-values (lambdas and weights are initialized with standard values)
		gdist->init(r);

		// fit the distribution
		fit_base(gdist);

		TGED ged;
		ged.loglikelihood = gdist->compute_loglikelihood(trace);
		ged.m = gdist->m;
		ged.r = gdist->r;
		ged.lambdas = gdist->lambdas;
		ged.weights = gdist->weights;

		queue->insert(ged);

		cout << ged.loglikelihood << endl;
	}
	else
	{
		for (unsigned int i = k; i <= N / m; i++)
		{
			r[pos] = i;

			// set r_pos+1 approprite with given r_pos = i
			fit_settings(m - 1, N - i, i, r, pos + 1, gdist);
		}
	}
}


//-----------------------------------------------------------------------------
// Calculate number of possibilities to chose the number of phases r_1,...,r_m
// of the hyper-Erlang distribution for a given number of Erlang
// branches m and a given overall number of states N.
//   m = number of Erlang branches
//   N = overall number of states
//   k = use at least Erlang phases of length k, i.e., 
//       w.l.o.g.: k <= r_1 <= r_2 <= ... <= r_m
//-----------------------------------------------------------------------------
unsigned long TGerlangFit::calc_num_settings(unsigned int m, unsigned int N, unsigned int k)
{
	if (m == 1) return 1;

	unsigned long _count = 0;

	for (unsigned int i = k; i <= N / m; i++)
	{
		_count += calc_num_settings(m - 1, N - i, i);
	}

	return _count;
}


unsigned long neighbor_settings(unsigned int* u, unsigned int m, unsigned int N, unsigned int k, unsigned int* r, unsigned int pos)
{
	if (m == 1)
	{
		unsigned int distance = 0;
		r[pos] = N;
		for (unsigned int i = 0; i < pos+1; i++)
		{
			if (r[i] > u[i]) distance += r[i]-u[i];
			else distance += u[i]-r[i];
		}
		if (distance == 2) return 1;
		else return 0;
	}

	unsigned long _count = 0;

	for (unsigned int i = k; i <= N / m; i++)
	{
		r[pos] = i;

		// set r_pos+1 approprite with given r_pos = i
		_count += neighbor_settings(u, m - 1, N - i, i, r, pos + 1);
	}

	return _count;
}

//------------------------------------------------------------------------
// Enumerate all possibilities to chose the number of phases r_1,...,r_m
// of the hyper-Erlang distribution for a given number of Erlang
// branches m and a given overall number of states N.
//------------------------------------------------------------------------
unsigned long TGerlangFit::enum_settings(unsigned int m, unsigned int N, unsigned int k, unsigned int* r, unsigned int pos)
{
	if (m == 1)
	{
		// print the setting of r_i's
		for (unsigned int i = 0; i < pos; i++)
		{
			if (r[i] > 0) cout << r[i] << ", ";
		}
		r[pos] = N;
		cout << r[pos];
		unsigned int* u = new unsigned int[pos+1];
		cout << " #neighbors = " << neighbor_settings(r, pos+1, pos+1, 0, u, 0) << endl;
		delete [] u;
		return 1;
	}

	unsigned long _count = 0;

	for (unsigned int i = k; i <= N / m; i++)
	{
		r[pos] = i;

		// set r_pos+1 approprite with given r_pos = i
		_count += enum_settings(m - 1, N - i, i, r, pos + 1);
	}

	return _count;
}
